
import java.util.*;

public final class Bot {
   
   public static Hashtable allBots = new Hashtable();
   
   public String id;
   public Point loc = new Point(0,0);
   public boolean alive = false;
   public Hashtable myBoxs = new Hashtable();  //mix of actual and pseudo boxs
   public int load = 0;
   public int score = 0;
   
   public Bot(String id) {
      this.id = id;
      allBots.put(id, this);
//System.err.println("Bot " + id + " created"); //DEBUG
   }

   public final void move(char dir) {
      switch (dir) {
         case 'N':
            loc.y++;
            break;
         case 'S':
            loc.y--;
            break;
         case 'E':
            loc.x++;
            break;
         case 'W':
            loc.x--;
            break;
      }
//System.err.println("Bot " + id + " moved " + dir + " to " + loc); //DEBUG
   }

   public final void setXY(int x, int y) {
      loc.x = x;
      loc.y = y;
//System.err.println("Bot " + id + " moved to " + loc); //DEBUG
   }

   public final Point closestDestination() {
      int minDist = 2000;  //longest distance between pts on largest board
      Point result = null;
      for (Iterator i = myBoxs.values().iterator(); i.hasNext(); ) {
         Box h = (Box) i.next();
         int dist = loc.distance(h.dest);
         if (dist < minDist) {
            result = h.dest;
            minDist = dist;
         }
      }
      return result;
   }

   public static Vector closest = new Vector();

   public final int closestBot() {
      closest.clear();
      int minDist = 2000;  //longest distance between pts on largest board
      for (Iterator i = allBots.values().iterator(); i.hasNext(); ) {
         Bot r = (Bot) i.next();
         if (r == this) continue;
         int dist = loc.distance(r.loc);
         if (dist < minDist) {
            if (dist == 1) closest.add(r);
            minDist = dist;
         }
      }
      return minDist;
   }

   public final String getDropList() {
      StringBuffer result = new StringBuffer();
      for (Iterator i = myBoxs.values().iterator(); i.hasNext(); ) {
         Box h = (Box) i.next();
         if (loc.equals(h.dest)) {
            result.append(h.id);
            result.append(' ');
         }
      }
      return result.toString();
   }

   public final void pickup(String pkg) {
      Box p = (Box) Box.freeBoxs.get(pkg);
      if (p == null) {  //didn't pick up a known free box
         p = Box.pseudoBox(pkg, loc); //this box will change loc w/ the robot
      }
      else {
         p.untrack();
      }
      load += p.weight;
      p.held = true;
      myBoxs.put(pkg, p);
//System.err.println("Bot " + id + " picked up pkg " + p.id + " at " + loc); //DEBUG
   }

   public final void drop(String pkg) {
      //this could have been a destination and the boxd disappeared from the game
      Point pt = new Point(loc.x, loc.y);
      Box p = (Box) Box.heldBoxs.remove(pkg); //dropped so no longer held
      if (p == null) {
         p = Box.pseudoBox(pkg, pt); 
//System.err.println("Bot " + id + " dropped pseudo pkg " + p.id + " at " + loc); //DEBUG
      }
      else {
         if (loc.equals(p.dest)) { //if we know the dest, we know the weight
            score += p.weight;
            p.disappear();
//System.err.println("Bot " + id + " dropped actual pkg " + p.id + " at destination " + loc); //DEBUG
         }
         else {
            p.loc = pt; //update location since we were not tracking
            p.track();
//System.err.println("Bot " + id + " dropped actual pkg " + p.id + " at " + loc); //DEBUG
         }
      }
      p.held = false;
      load -= p.weight;
      myBoxs.remove(pkg);
   }
   
   public static void cleanUp() {
      for (Iterator i = allBots.keySet().iterator(); i.hasNext(); ) {
         Bot r = (Bot) allBots.get(i.next());
         if (!r.alive) {
//System.err.println("Bot " + r.id + " has died");  //DEBUG
            //delete all boxs held by this robot, both pseudo and held
            //no held box should be currently tracking
            for (Iterator j = r.myBoxs.keySet().iterator(); j.hasNext(); ) {
               Box p = (Box) r.myBoxs.get(j.next());
               p.disappear();  //disappear the box
               j.remove();
            }
            i.remove();
         }
      }
   }

}