
import java.util.*;

public final class Box implements Comparable {

   public static Hashtable heldBoxs = new Hashtable();
   public static Hashtable freeBoxs = new Hashtable();
   public static Hashtable pkgsByLocation = new Hashtable();
   public static Hashtable pseudoBoxs = new Hashtable();
   
   public String id;
   public Point loc;
   public Point dest;  //null -> pseudoBox
   public int weight;
   public boolean held = false;
   public boolean tracking = false;
   
   public Box(String id) {
      this.id = id;
   }

   //mark this box for retrieval by location since it is available
   //for pickup
   public final void track() {
      if (tracking) return;  //already tracking
      if (loc == null) return; //can't track if we don't know the loc
      TreeSet h1 = (TreeSet) pkgsByLocation.get(loc);
      if (h1 == null) {
         h1 = new TreeSet();
         pkgsByLocation.put(loc, h1);
      }
      h1.add(this);
      tracking = true;
      freeBoxs.put(id, this);
   }
   
   //do not follow closely the locations of this box as it is
   //being held by a robot
   public final void untrack() {
      if (!tracking) return; //can't be tracking this one
      if (loc == null) return; //can't be tracking this one
      TreeSet h1 = (TreeSet) pkgsByLocation.get(loc);
      if (h1 != null) {
         h1.remove(this);
         if (h1.isEmpty()) pkgsByLocation.remove(loc);
      }
      tracking = false;
      freeBoxs.remove(id);
      heldBoxs.put(id, this);
   }
   
   public static final void reportedBox(String id, int dx, int dy, int w, Point l) {
      Box p = null;
      if (!freeBoxs.containsKey(id)) { //if this is a new box
         if ((p = (Box)pseudoBoxs.get(id)) != null) {
            //here we stumbled across a dropped box that we
            //previously saw picked up by another robot but did
            //not know the details of
//System.err.println("Former pseudo box found: " + id);
            pseudoBoxs.remove(id);
         }
         else {
            p = new Box(id);
         }
         p.weight = w;
         p.dest = new Point(dx, dy);
         p.loc = new Point(l.x, l.y);
         p.track();  //this will add it to freeBoxs
      }
   }
   
   public static final Box pseudoBox(String id, Point l) {
      Box p = (Box) pseudoBoxs.get(id);
      if (p == null) {
         p = new Box(id);
         pseudoBoxs.put(id, p);
      }
      p.loc = l;
      return p;
   }

   public final boolean equals(Object pkg) {
      try {
         if (id.equals(((Box)pkg).id)) return true;
      } catch (Exception ex) {
      }
      return false;
   }

   public final int compareTo(Object pkg) {
      if (equals(pkg)) return 0;
      Box p = (Box)pkg;
      if (weight == p.weight) {
         int d1 = loc.distance(dest);
         int d2 = p.loc.distance(p.dest);
         if (d1 == d2) {   
            return -id.compareTo(p.id);
         }
         else {
            return d1 - d2; //ascending on distance to dest
         }
      }
      return p.weight - weight; //this puts weights in descending order
   }
   
   public static final Iterator getBoxList(Point loc) {
      TreeSet t = (TreeSet) pkgsByLocation.get(loc);
      return t == null ? null : t.iterator();
   }
   
   public final void disappear() {
      //only boxs that were held can possibly disappear
      //we are not tracking any box that was held
      heldBoxs.remove(id);
      freeBoxs.remove(id); //should never be here, but just in case
      pseudoBoxs.remove(id);
   }
   
   //closest box being held, including pseudo boxs
   public static Point closestHeldBox(Bot r) {
      int minDist = 2001;  //longest distance between pts on largest board
      Point result = null;
      for (Iterator i = Bot.allBots.values().iterator(); i.hasNext(); ) {
         Bot b = (Bot) i.next();
         if (b == r) continue;
         if (b.myBoxs.isEmpty()) continue;
         int dist = r.loc.distance(b.loc);
         if (dist < minDist) {
            result = b.loc;
            minDist = dist;
         }
      }
      return result;
   }
   
   //this will be a dropped pseudo box because closestHeldBox
   //handles held pseudo boxs
   public static Point closestPseudoBox(Point p) {
      int minFreeDist = 2001;  //longest distance between pts on largest board
      Point freeResult = null;
      for (Iterator i = pseudoBoxs.values().iterator(); i.hasNext(); ) {
         Box pkg = (Box)i.next();
         if (!pkg.held) {
            Point h = pkg.loc;
            int dist = p.distance(h);
            if (dist < minFreeDist) {
               freeResult = h;
               minFreeDist = dist;
            }
         }
      }
      return freeResult;
   }
}