
import java.util.*;

public final class GameBoard {

   public static final char BASE = '@';
   public static final char WALL = '#';
   public static final char LETHAL = '~';
   public static final char SAFE = '.';

   public static Vector baseList = new Vector();

   public static int numRows;
   public static int numCols;
   public static char[][] theBoard;
   public static int[] theBits;
   public static int topPos;
   public static int bottomPos;
   public static int maxPos;
   public static boolean top = true;
   public static Point search;

   public static int[] masks = new int[32];
   
   static {
      int mask = 1;
      for (int i = 0; i < 32; i++) {
         masks[i] = mask;
         mask <<= 1;
      }
   }

   public GameBoard(int rows, int cols) {
      numRows = rows;
      numCols = cols;
      theBoard = new char[rows][cols];
      topPos = rows * cols - 1;
      maxPos = topPos + 1;
      theBits = new int[(maxPos + 32) / 32];
      search = new Point(numCols, numRows);
   }
   
   public static void setRow(int row, String data) {
      data.getChars(0, numCols, theBoard[row], 0);
      for (int i = 0; i < numCols; i++) {
         switch (theBoard[row][i]) {
            case BASE:
               baseList.add(new Point(i + 1, row + 1));
               break;
            case WALL: case LETHAL:
               int idx = (row * numCols + i) >>> 5;
               int bit = (row * numCols + i) & 0x1F;
               theBits[idx] |= masks[bit];
               break;
         }
      }
   }
   
   public static int canKill(Point from, char dir) {
      int count = 0;
      try {
         int x = from.x;
         int y = from.y;
         while (true) {
            switch (dir) {
               case 'E': x++;
                  break;
               case 'W': x--;
                  break;
               case 'N': y++;
                  break;
               case 'S': y--;
                  break;
            }
            count++;
            if (theBoard[y - 1][x - 1] == LETHAL) return count;
            if (theBoard[y - 1][x - 1] == WALL) return 0;
         }
      } catch (Exception ex) {
         return 0;  //array index out of bounds == edge of world == WALL
      }
   }

   public static void mark(Point p) {
      int loc = (p.y - 1) * numCols + (p.x - 1);
      int idx = loc >>> 5;  // divide by 32
      int bit = loc & 0x1F; // mod 32
      theBits[idx] |= masks[bit];  //mark as visited
      if (baseList.contains(p)) { //if it was a base, mark as SAFE
         baseList.remove(p);
         theBoard[p.y - 1][p.x - 1] = SAFE;
      }
   }
   
   public static boolean visited(Point p) {
      int loc = (p.y - 1) * numCols + (p.x - 1);
      int idx = loc >>> 5;  // divide by 32
      int bit = loc & 0x1F; // mod 32
      return (theBits[idx] & masks[bit]) != 0;  //mark as visited
   }
   
   public static Point nextSearch() {
      if (!visited(search)) return search;
      if (top) {
         while (topPos >= 0) {
            int idx = topPos >>> 5;  // divide by 32
            int bit = topPos & 0x1F; // mod 32
            if ((theBits[idx] & masks[bit]) == 0) { 
               //never been here
               top = false;
               search = new Point(topPos % numCols + 1, topPos / numCols + 1);
               return search;
            }
            topPos--;
         }
      }
      else {
         while (bottomPos < maxPos) {
            int idx = bottomPos >>> 5;  // divide by 32
            int bit = bottomPos & 0x1F; // mod 32
            if ((theBits[idx] & masks[bit]) == 0) { 
               //never been here
               top = true;
               search = new Point(bottomPos % numCols + 1, bottomPos / numCols + 1);
               return search;
            }
            bottomPos++;
         }
      }
      return null;
   }

   public static char getTile(int x, int y) {
      try {
         return theBoard[y - 1][x - 1];
      } catch (Exception ex) {
         return WALL;
      }
   }
   
}