/*
 * This is the main class for the ICFP competition.  It represents a 
 * robot controller in a package delivery competition
 * 
 * 2002 entry by Chris Eagle cseagleATredshiftDOTcom
 */

import java.io.*;
import java.net.*;
import java.util.*;

public final class SunsRevenge {

   public static Socket sock;
   public static InputStream is;
   public static OutputStream os;
   public static BufferedReader br;
   public static Bot me;
   public static int capacity;
   public static int load;
   public static int money;
   public static int pickTries;
   public static boolean newBoxs = false;
   
   public static char desiredMove;
   public static int badCount;

   public static final void sendMove(char mv, String move) throws Exception {
      desiredMove = mv;
      os.write(move.getBytes());
   }

   public static boolean reachable(Point p) {
      return me.loc.distance(p) < money;
   }
   
   public static void readBoard() throws Exception {
      StringTokenizer st = new StringTokenizer(br.readLine());
      int cols = Integer.parseInt(st.nextToken());
      int rows = Integer.parseInt(st.nextToken());
      new GameBoard(rows, cols);
      for (int i = 0; i < rows; i++) {
         GameBoard.setRow(i, br.readLine());
      }
//System.err.println("GameBoard\n" + GameBoard.boardString()); //DEBUG
   }
   
   public static void readConfig() throws Exception {
      StringTokenizer st = new StringTokenizer(br.readLine());
      me = new Bot(st.nextToken());
      capacity = Integer.parseInt(st.nextToken());
      money = Integer.parseInt(st.nextToken());
//System.err.println("Playing as robot " + me.id + " with $" + money + //DEBUG
//                   " and cap = " + capacity); //DEBUG
   }

   public static void readBoxs() throws Exception {
      String input = br.readLine();
      StringTokenizer st = new StringTokenizer(input);
      newBoxs = st.hasMoreTokens();
      if (newBoxs) {
         while (st.hasMoreTokens()) {
            String id = st.nextToken();
            int dx = Integer.parseInt(st.nextToken());
            int dy = Integer.parseInt(st.nextToken());
            int w = Integer.parseInt(st.nextToken());
            Box.reportedBox(id, dx, dy, w, me.loc);
         }
      }
      for (Iterator i = Box.pseudoBoxs.values().iterator(); i.hasNext(); ) {
         Box p = (Box) i.next();
         if (p.loc.equals(me.loc)) { 
            //if there is a pseudo box here, it must not have been
            //reported above, and therefore must have disappeared, so remove it
            i.remove();
         }
      }
   }

   public static void readResponse() throws Exception {
      Bot bot = null;
      StringTokenizer st = null;
      int x = 0;
      for (Iterator i = Bot.allBots.values().iterator(); i.hasNext(); ) {
         ((Bot)i.next()).alive = false;
      }
      try {
         st = new StringTokenizer(br.readLine());
      } catch (Exception ex) {
         //something went horrible wrong, we can't read the last response
         //and our state is completely out of whack, so just die
         sendMove('X', "1 Die\n"); //invalid command should kill me off
         return;
      }
      while (st.hasMoreTokens()) {
         String tok = st.nextToken();
         char ch = tok.charAt(0);
         if (ch == '#') {
            String theBot = tok.substring(1);
            bot = (Bot) Bot.allBots.get(theBot);
            if (bot == null) {
               bot = new Bot(theBot);
            }
            bot.alive = true;
         }
         else {
            if (bot == me && desiredMove != ch) {
               badCount++;
            }
            else {
               badCount = 0;
            }
            switch (ch) {
               case 'P':
                  bot.pickup(st.nextToken());
                  break;
               case 'D':
                  bot.drop(st.nextToken());
                  break;
               case 'X':
                  x = Integer.parseInt(st.nextToken());
                  break;
               case 'Y':
                  if (bot == me) badCount -= 2;
                  bot.setXY(x, Integer.parseInt(st.nextToken()));
                  break;
               default: //must be a move
                  bot.move(ch);
                  break;
            }
         }
      }
      Bot.cleanUp();
   }

   public static void init() throws Exception {
      os.write("Player\n".getBytes());
      readBoard();
      readConfig();
      readResponse();
   }

   public static PriorityNode AStarSearch(Point start, Point goal) {
      PriorityQueue open = new PriorityQueue();
      Hashtable closed = new Hashtable();
      PriorityNode n = new PriorityNode(start, null, goal);
      open.add(n);
      Vector v = new Vector();
      while (!open.isEmpty()) {
         n = open.popFirst();
         if (n.equals(goal)) {
            return n;
         }
         v.clear();
         v.add(new PriorityNode(n.x - 1, n.y, n, goal));
         v.add(new PriorityNode(n.x + 1, n.y, n, goal));
         v.add(new PriorityNode(n.x, n.y + 1, n, goal));
         v.add(new PriorityNode(n.x, n.y - 1, n, goal));
         for (int i = 0; i < 4; i++) {
            PriorityNode np = (PriorityNode) v.elementAt(i);
            PriorityNode temp = null;
            char ch = GameBoard.getTile(np.x, np.y);
            if (ch == GameBoard.SAFE || ch == GameBoard.BASE) {
               if ((temp = open.find(np)) != null) {
                  if (temp.g <= np.g) continue;
                  open.remove(temp);
               } else if ((temp = (PriorityNode)closed.get(np.toString())) != null) {
                  if (temp.g <= np.g) continue;
                  closed.remove(np.toString());
               }
               open.add(np);
            }
         }
         closed.put(n.toString(), n);
      }
      return null; // if no path found 
   }
   
   public static final char getMoveFromNode(PriorityNode n) {
      while (n != null) {
         if (n.prev != null && n.prev.prev == null) {
            char mv = 'N';
            if (n.x == me.loc.x) {
               return n.y > me.loc.y ? 'N' : 'S';
            }
            else {
               return n.x > me.loc.x ? 'E' : 'W';
            }
         }
         n = n.prev;
      }
      return 'X';
   }
   
   //closest box not being held
   public static char moveForNearestBox() {
      int cost = 1000000;
      PriorityNode resultNode = null;
      for (Iterator i = Box.pkgsByLocation.keySet().iterator(); i.hasNext(); ) {
         Point p = (Point) i.next();
         PriorityNode n = AStarSearch(me.loc, p);
         if (n != null) {
            if (n.g < cost) {
               cost = n.g;
               resultNode = n;
            }
         }
      }
      return getMoveFromNode(resultNode);
   }
   
   public static final char moveForNearestHome() {
      int cost = 1000000;
      PriorityNode resultNode = null;
      for (Iterator i = GameBoard.baseList.iterator(); i.hasNext(); ) {
         Point p = (Point) i.next();
         PriorityNode n = AStarSearch(me.loc, p);
         if (n != null) {
            if (n.g < cost) {
               cost = n.g;
               resultNode = n;
            }
         }
      }
      return getMoveFromNode(resultNode);
   }
   
   public static boolean useMove(char mv) throws Exception {
      if (mv == 'X') return false;
      int myBid = badCount == 0 ? 1 : badCount * 5; 
      if (myBid > money) myBid = money;
      sendMove(mv, myBid + " Move " + mv + "\n");
      money -= myBid;
//System.err.println("Moving : " + mv);  //DEBUG
      return true;
   }
   
   public static boolean moveTowards(Point goal) throws Exception {
      if (goal == null) return false;
      //move towards closest box or base
//System.err.println("Goal is: " + goal);  //DEBUG
      //look for shortest path to various homes
      PriorityNode n = AStarSearch(me.loc, goal);
      return useMove(getMoveFromNode(n));
   }
   
   public static boolean canDrop() throws Exception {
      //drop the right boxs
      String list = me.getDropList();
      if (list.length() > 0) {
         int myBid = closestBot == 1 ? 10 : 1;
         if (myBid > money) myBid = money;
         sendMove('D', myBid + " Drop " + list + "\n");
         money -= myBid;
//System.err.println("Dropping : " + list);  //DEBUG
         return true;
      }
      return false;
   }
   
   public static boolean canPick() throws Exception {
      //pick up some boxs
      Iterator i = Box.getBoxList(me.loc);
      int pickCount = 0;
      if (i != null) {
         StringBuffer list = new StringBuffer();
         int desiredLoad = me.load;
         while (i.hasNext()) {
            Box p = (Box)i.next();
            if (p.held) continue; //held by me
            if (p.weight <= (capacity - desiredLoad)) {
               desiredLoad += p.weight;
               list.append(p.id);
               list.append(' ');
               pickCount++;
            }
         }
         if (list.length() > 0) {
            list.append('\n');
            int myBid = badCount == 0 ? 1 : badCount * 5; 
            if (myBid > money) myBid = money;
//System.out.println("Bidding " + myBid + " for this pickup");  //DEBUG
//QUESTION - determine if other robots are near, and if likely to get bumped
//           bid higher so you can pick before getting bumped, if list length
//           is only one, you are going to get bumped and drop it anyway so
//           bid only 1            
            sendMove('P', myBid + " Pick " + list);
            money -= myBid;
//System.err.println("Picking : " + list);  //DEBUG
            return true;
         }
      }
      return false; //no boxs I can pick up
   }
   
   public static boolean canDeliver() throws Exception {
      //move towards destination - change to nearest
      return moveTowards(me.closestDestination());
   }
   
   public static final boolean moveToPseudoBox() throws Exception {
      //move towards destination - change to nearest
      return moveTowards(Box.closestPseudoBox(me.loc));
   }
   
   public static final void randomMove() throws Exception {
      if (!moveTowards(GameBoard.nextSearch())) {
         sendMove('P', "1 Pick\n");
         money--;
      }
   }
   
   public static int closestBot;
   public static Vector tooClose;
   
   public static void getAdjacentBotList() {
      closestBot = me.closestBot();
      tooClose = Bot.closest;
   }
   
   public static char whichWay(Bot r1, Bot r2) {
      if (r2.loc.x > r1.loc.x) return 'E';
      if (r2.loc.x < r1.loc.x) return 'W';
      if (r2.loc.y < r1.loc.y) return 'S';
      return 'N';
   }
   
   public static boolean killMode = false;

   public static boolean canKill() throws Exception {
      if (closestBot == 1) {
         for (Iterator i = tooClose.iterator(); i.hasNext(); ) {
            Bot r = (Bot) i.next();
            if (GameBoard.getTile(r.loc.x, r.loc.y) == GameBoard.LETHAL) {
               continue;//not sure why I need to do this, but I do
            }
            char dir = whichWay(me, r);
            int dist = GameBoard.canKill(me.loc, dir);
            if (dist > 0) {
               int bid;
               if (!killMode && dist == 1 && money > 250) {
                  bid = 250; 
               }
               else if (money >= dist * 50) {
                  bid = 50;
               }
               else {
                  bid = money / dist;
               }
               bid = bid > 0 ? bid : 1;
               sendMove(dir, bid + " Move " + dir + "\n");
               money -= bid;
               killMode = true;
               return true;
            }
         }
      }
      killMode = false;
      return false;
   }
   
   public static boolean mustProtect() throws Exception {
      if (closestBot == 1) {
         for (Iterator i = tooClose.iterator(); i.hasNext(); ) {
            Bot r = (Bot) i.next();
            if (GameBoard.getTile(r.loc.x, r.loc.y) == GameBoard.LETHAL) {
               continue;//not sure why I need to do this, but I do
            }
            char dir = whichWay(me, r);
            char tile = GameBoard.SAFE;
            switch (dir) {
               case 'E':
                  tile = GameBoard.getTile(me.loc.x - 1, me.loc.y);
                  break;
               case 'W':
                  tile = GameBoard.getTile(me.loc.x + 1, me.loc.y);
                  break;
               case 'S':
                  tile = GameBoard.getTile(me.loc.x, me.loc.y + 1);
                  break;
               case 'N':
                  tile = GameBoard.getTile(me.loc.x, me.loc.y - 1);
                  break;
            }   
            if (tile == GameBoard.LETHAL) { //I can get pushed
               int bid = 1;
               if (money > 250) {
                  bid = 250;
               }
               else {
                  bid = money > 50 ? 50 : 1;
               }
               sendMove(dir, bid + " Move " + dir + "\n"); //push the threat
               money -= bid;
               return true;
            }
         }
      }
      return false;
   }

   public static void move() throws Exception {
      //need a measure of safety, check distance to other bots
      //and alter behavior if we are too close
      getAdjacentBotList(); //setup for mustProtect and canKill
      if (mustProtect()) return;
      if (canDrop()) return;
      if (canKill()) return;
      if (canPick()) return; //adjust this so don't get in a pick war if I already have boxs
      if (canDeliver()) return;
      if (useMove(moveForNearestBox())) return;
      //QUESTION - need to do better here, don't go for closest home
      //go for the one we can reach fastest 
      if (useMove(moveForNearestHome())) return;
      if (moveTowards(Box.closestHeldBox(me))) return;
      //no good goal, so see if there are known pseudoBoxs
      if (moveTowards(Box.closestPseudoBox(me.loc))) return;
      //QUESTION - need to do a random move in search of dropped 
      //boxs
      //should never get here in a real game because we should have
      //seen all pacakges picked up and dropped and at least see them
      //in the pseudoBoxs list
      randomMove(); //maintain bitmap
   }

   public static void play() throws Exception {
      int turn = 1;
      while (Bot.allBots.containsKey(me.id)) {
         newBoxs = false;
         readBoxs();
//System.err.println("Turn = " + turn++); //DEBUG         
         move();
         readResponse();
         GameBoard.mark(me.loc);  //do this here because there is no better place
      }
//      System.err.println("Damn I died!");  //DEBUG
   }
   
   public static void main(String args[]) {
      try {
         if (args.length != 2) {
            System.err.println("usage error, not enough arguments");
         }
         else {
            sock = new Socket(args[0], Integer.parseInt(args[1]));
            is = sock.getInputStream();
            os = sock.getOutputStream();
            br = new BufferedReader(new InputStreamReader(is));
//System.err.println("Connected");
            init();
            play();
         }
      } catch (Exception ex) {
//System.err.println("Final score should be: " + me.score + ", currency = " + money);
//         ex.printStackTrace();  //DEBUG
         try {
            sock.close();  //be nice and close socket
         } catch (Exception cex) {
         }
      }
   }

}